cdef class BitByteBuffer(object):
    cdef unsigned char[:] ba
    cdef int lastbyte
    cdef int lastbits
    cdef int cnt
    cdef int lba
    cdef int linc
    
    def __init__(self, unsigned char[:] byts):
        self.ba = byts
        self.lastbyte = 0
        self.lastbits = 0
        self.cnt = 0
        self.lba = self.ba.shape[0]
        self.linc = self.lba // 5
        
    def receiveint(self, int nbits):
        cdef int num = 0
        cdef int mask = (1<< nbits) -1
        cdef int lastbyte = self.lastbyte
        cdef int lastbits = self.lastbits
        cdef int cnt = 0
        cdef unsigned char[:] ba = self.ba[self.cnt:self.cnt+32]
        
        while nbits >= 8:
            lastbyte = ( lastbyte << 8) | ba[cnt]
            cnt += 1
            num |= (lastbyte >> lastbits) << (nbits - 8)
            nbits -= 8
        if nbits > 0:
            if lastbits < nbits:
                lastbits += 8
                lastbyte = (lastbyte << 8) | ba[cnt]
                cnt += 1
            lastbits -= nbits
            num |= (lastbyte >> lastbits) & ((1 << nbits) -1)
        num &= mask
        lastbyte &= 0xff
        self.lastbyte = lastbyte
        self.lastbits = lastbits
        self.cnt += cnt
        return num
    
    def sendint(self, int nbits, int num):
        
        cdef int nbytes = (nbits + 7) // 8 + 1
        cdef int cnt = 0
        cdef int lastbyte = self.lastbyte
        cdef int lastbits = self.lastbits
        cdef unsigned char[32] ba 
        cdef int j
        while nbits >= 8:
            lastbyte = (lastbyte << 8) | (num >> (nbits - 8))
            ba[cnt] = (lastbyte >> lastbits) & 0xff
            cnt += 1
            nbits -= 8
        if nbits > 0:
            lastbyte = (lastbyte << nbits) | num
            lastbits += nbits
            if lastbits >= 8:
                lastbits -= 8
                ba[cnt] = (lastbyte >> lastbits) & 0xff
                cnt += 1
        if lastbits > 0:
            ba[cnt] = (lastbyte << (8 - lastbits)) & 0xff
        lastbyte &= 0xff
        for j in range(cnt + 1):
            self.ba[self.cnt + j] = ba[j]
        self.cnt += cnt
        self.lastbyte = lastbyte
        self.lastbits = lastbits

    def sendbyte(self, int nbits, unsigned char  num):
        
        cdef int cnt = 0
        cdef int lastbyte = self.lastbyte
        cdef int lastbits = self.lastbits

        cdef unsigned char[2] ba 
        cdef int j
        if nbits > 0:
            lastbyte = (lastbyte << nbits) | num
            lastbits += nbits
            if lastbits >= 8:
                lastbits -= 8
                ba[cnt] = (lastbyte >> lastbits) & 0xff
                cnt += 1
        if lastbits > 0:
            ba[cnt] = (lastbyte << (8 - lastbits)) & 0xff
        lastbyte &= 0xff
        self.ba[self.cnt] = ba[0]
        if cnt == 1:
            self.ba[self.cnt + 1] = ba[1]
        self.cnt += cnt
        self.lastbyte = lastbyte
        self.lastbits = lastbits

    def rewind(self):
        self.lastbits = 0
        self.lastbyte = 0
        self.cnt = 0
        
    def tobytes(self):
        return bytearray(self.ba[:self.cnt + 1])
    
    def receiveints(self, int bitsize, long[:] sizes):
        cdef int i, j, nbytes, p, num
        cdef int nums[3]
        cdef int ba[32]
        ba[0] = ba[1] = ba[2] = ba[3] = 0
      
        nbytes = 0
        while bitsize > 8:
            ba[nbytes] = self.receiveint(8)
            nbytes += 1
            bitsize -= 8
        if bitsize > 0:
            ba[nbytes] = self.receiveint(bitsize)
            nbytes += 1
        for i in range(2, 0, -1):
            num = 0
            for j in range(nbytes-1, -1, -1):
                num = (num << 8) | ba[j]
                p = num // sizes[i]
                ba[j] = p
                num = num - p * sizes[i]
            nums[i] = num
        nums[0] = ba[0] | (ba[1] << 8) | (ba[2] << 16) | (ba[3] << 24)
        return nums

    def sendints(self, int bitsize, long[:] sizes, int[:] nums):
        cdef unsigned char[32] ba
        cdef int nbytes = 0
        cdef long bignum
        cdef int i
        bignum = (((nums[0] * sizes[1]) + nums[1]) * sizes[2]) + nums[2]

        while bignum > 0:
            ba[nbytes] = bignum & 0xff
            nbytes += 1
            bignum >>= 8
            
        if bitsize >= nbytes * 8:
            for i in range(nbytes):
                self.sendbyte(8, ba[i])
            self.sendbyte(bitsize - nbytes * 8, 0)
        else:
            for i in range(nbytes-1):
                self.sendbyte(8, ba[i])
            self.sendbyte(bitsize - (nbytes-1) * 8, ba[nbytes-1])
    
def unpack_crds(bs, ms, long maxatoms, long natoms, long bitsize, 
                long [:] bitsizeint, long [:] sizes, long [:] minint):
    """
    Unpack a set of coordinates from a BitByteBuffer
    """
    import numpy as np
    
    cdef int run, flag, is_smaller, j, k
    cdef Py_ssize_t i
    
    i = 0
    run = 0
    crds = np.zeros((maxatoms, 3), dtype=np.int)
    cdef long [:,:] crds_v = crds
    
    cdef int smallnum, smallidx
    sizesmall = np.zeros((3), dtype=np.int)
    
    cdef long nums[3]
    cdef long thiscoord[3]
    cdef long prevcoord[3]
    cdef long tmp[3]
    
    cdef long minint0 = minint[0]
    cdef long minint1 = minint[1]
    cdef long minint2 = minint[2]
    
    
    while i < natoms:
        if bitsize == 0:
            thiscoord[0] = bs.receiveint(bitsizeint[0])
            thiscoord[1] = bs.receiveint(bitsizeint[1])
            thiscoord[2] = bs.receiveint(bitsizeint[2])
            
        else:
            thiscoord = bs.receiveints(bitsize, sizes)
        thiscoord[0] += minint0
        thiscoord[1] += minint1
        thiscoord[2] += minint2
        prevcoord = thiscoord

        flag = bs.receiveint(1)
        is_smaller = 0
        if flag == 1:
            run = bs.receiveint(5)
            is_smaller = run % 3
            run -= is_smaller
            is_smaller -= 1
        
        if run > 0:
            smallnum = ms.smallnum
            sizesmall = ms.sizesmall
            smallidx = ms.smallidx
            for k in range(0, run, 3):
                thiscoord = bs.receiveints(smallidx, sizesmall)
                thiscoord[0] += prevcoord[0] - smallnum
                thiscoord[1] += prevcoord[1] - smallnum
                thiscoord[2] += prevcoord[2] - smallnum
                if (k == 0):
                    tmp = thiscoord
                    thiscoord = prevcoord
                    prevcoord = tmp
                    crds_v[i, 0] = prevcoord[0]
                    crds_v[i, 1] = prevcoord[1]
                    crds_v[i, 2] = prevcoord[2]
                    i += 1
                else:
                    prevcoord = thiscoord
                crds_v[i, 0] = thiscoord[0]
                crds_v[i, 1] = thiscoord[1]
                crds_v[i, 2] = thiscoord[2]
                i += 1
        else:
            crds_v[i, 0] = thiscoord[0]
            crds_v[i, 1] = thiscoord[1]
            crds_v[i, 2] = thiscoord[2]
            i += 1
        
        ms.update(is_smaller)

    return crds



import numpy as np

def cdist(XA, XB):
  if len(XA.shape) != 2:
    raise ValueError('Error: XA must be a 2D array')
  if len(XB.shape) != 2:
    raise ValueError('Error: XB must be a 2D array')
  ma, na = XA.shape
  mb, nb = XB.shape
  if (na != nb):
    raise ValueError('Error: XA and XB must have the same second dimension')

  cd = np.zeros((ma, mb))

  for ia, va in enumerate(XA):
    dx = XB - va
    r = np.sqrt((dx * dx).sum(axis=1))
    cd[ia] = r

  return cd



import numpy as np
import mdio.base 
from mdio.utilities import la2v, v2la, selection_to_indices

def pdb_atom_parse(line):
    """
    Extract data from an ATOM or HETATM line.
    """
    if '\n' in line:
        line = line[:line.index('\n')]
    line = line.ljust(80)
    result = {}
    result['serial'] = int(line[6:11])
    result['name'] = line[12:16]
    result['altLoc'] = line[16]
    result['resName'] = line[17:20]
    result['chainID'] = line[21]
    result['resSeq'] = int(line[22:26])
    result['iCode'] = line[26]
    result['x'] = float(line[30:38]) * 0.1
    result['y'] = float(line[38:46]) * 0.1
    result['z'] = float(line[46:54]) * 0.1
    try:
        result['occupancy'] = float(line[54:60])
    except:
        result['occupancy'] = 1.0
    try:
        result['tempFactor'] = float(line[60:66])
    except:
        result['tempFactor'] = 0.0
    element = line[76:78].strip()
    if len(element) == 2:
        element = element[0] + element[1].lower()
    result['element'] = element
    
    if result['element'] == '':
        result['element'] = result['name'].lstrip(' 0123456789')[0]
    result['charge'] = line[78:80]
    return result
    
class PDBFileReader(object):
    def __init__(self, filename, top=None, selection=None, altloc='A'):
        #print('DEBUG: Reading header')
        self.filename = filename
        self.atom_indices = None
        self.selection = selection
        self.top = None
        self.altloc = altloc
        self.unitcell_vectors = None
        self.index = -1
        self.buffer = ' ' * 80
        self.f = open(filename, 'r')
        while not self.buffer[:6] in ['CRYST1', 'MODEL ', 'ATOM  ','HETATM', 'EOF   ']:
            self.buffer = self.f.readline()
            if self.buffer == "":
                  self.buffer = 'EOF   '
            else:
               self.buffer = self.buffer.ljust(80)
        if self.buffer[:6] == 'CRYST1':
            a = float(self.buffer[6:15]) * 0.1
            b = float(self.buffer[15:24]) * 0.1
            c =float(self.buffer[24:33]) * 0.1
            alpha = float(self.buffer[33:40])
            beta = float(self.buffer[40:47])
            gamma = float(self.buffer[47:54])
            self.unitcell_vectors = la2v((a, b, c), (alpha, beta, gamma))
        while not self.buffer[:6] in ['MODEL ', 'ATOM  ','HETATM', 'EOF   ']:
            self.buffer = self.f.readline()
            if self.buffer == "":
                self.buffer = 'EOF   '
            else:
                self.buffer = self.buffer.ljust(80)
        if self.buffer =='EOF   ':
            raise IOError('Error - this does not seem to be a valid PDB file.')
        #print('DEBUG: header read.')
            
    def read_frame(self):
        #print('DEBUG: reading frame.')
        if self.buffer[:6] == 'EOF   ':
            return None
        while not self.buffer[:6] in ['ATOM  ', 'HETATM', 'EOF   ']:
            self.buffer = self.f.readline()
            if self.buffer == "":
                self.buffer = 'EOF   '
            else:
                self.buffer = self.buffer.ljust(80)
        if self.buffer == 'EOF   ':
            return None

        result = []
        index = -1
        if self.buffer[:6] in ['ATOM  ', 'HETATM']:
            while self.buffer[:6] in ['ATOM  ', 'HETATM', 'TER   ']:
                if self.buffer[:6] in ['ATOM  ', 'HETATM']:
                    index += 1
                    data = pdb_atom_parse(self.buffer)
                    data['index'] = index
                    if self.altloc is None:
                        result.append(data)
                    else:
                        if data['altLoc'] in [self.altloc, ' ']:
                            data['altLoc'] = ' '
                            result.append(data)
                self.buffer = self.f.readline()
                if self.buffer == "":
                    self.buffer = 'EOF   '
                else:
                    self.buffer = self.buffer.ljust(80)
        if self.top is None:
            self.top = mdio.base.Topology(result)
            self.atom_indices = selection_to_indices(self.selection, self.top)
            if self.atom_indices is not None:
                self.top = self.top.subset(self.atom_indices)
        self.index += 1
        crds = np.array([[data['x'], data['y'], data['z']] for data in result])
        if self.atom_indices is not None:
            crds = crds[self.atom_indices]
        frame = mdio.base.Frame(crds, 
                      box=self.unitcell_vectors, 
                      time=self.index,
                      units='nanometers')
        #print('DEBUG: frame read.')

        return frame

    def read(self):
        frames = []
        frame = self.read_frame()
        while frame is not None:
            frames.append(frame)
            frame = self.read_frame()
        return mdio.base.Trajectory(frames, top=self.top)

    def close(self):
        self.f.close()
        
    def __enter__(self):
        return self
    
    def __exit__(self, *args):
        self.close()
        
class PDBFileWriter(object):
    def __init__(self, filename, top=None, use_variants=False):
        if top is None:
            raise RuntimeError('Error - cannot write PDB format without a topology.')
        if isinstance(top, mdio.base.Topology):
            self.top = top
        else:
            ttop = mdio.load(top)
            self.top = ttop.topology
        self.filename = filename
        self.index = -1
        self.use_variants = use_variants
        self.f = open(filename, 'w')

    def write_frame(self, frame):
        scalefactor = 10.0
        self.index += 1
        if self.index == 0:
            if frame.unitcell_vectors is not None:
                lengths, angles = v2la(frame.unitcell_vectors * scalefactor)
                crystformat = 'CRYST1{:9.3f}{:9.3f}{:9.3f}{:7.2f}{:7.2f}{:7.2f}\n'
                self.f.write((crystformat.format(lengths[0], lengths[1], lengths[2], angles[0], angles[1], angles[2])))
        modelformat = 'MODEL     {:4d}\n'
        self.f.write(modelformat.format(self.index + 1))
        atomformat = 'ATOM  {:5d} {:4s}{:1s}{:3s} {:1s}{:4d}{:1s}   {:8.3f}{:8.3f}{:8.3f}{:6.2f}{:6.2f}          {:2s}{:2s}\n'
        xyz = frame.xyz * scalefactor
        for i, atom in enumerate(self.top.atoms):
            name = atom.name.strip()
            if len(name) == 1:
                name = ' ' + name + '  '
            elif len(name) == 2:
                name = ' ' + name + ' '
            elif len(name) == 3:
                name = ' '  + name
            else:
                name = name[:4]
            if self.use_variants:
                if atom.residue.variant is not None:
                    resName = atom.residue.variant.strip().ljust(3)[:3]
                else:
                    resName = atom.residue.name.strip().ljust(3)[:3]
            else:
                resName = atom.residue.name.strip().ljust(3)[:3]
            chainID = atom.chain.chainid.strip().ljust(1)[0]
            self.f.write(atomformat.format(atom.serial,
                                           name, 
                                           atom.altloc,
                                           resName,
                                           chainID,
                                           atom.residue.resSeq,
                                           atom.residue.icode,
                                           xyz[i,0],
                                           xyz[i,1],
                                           xyz[i,2],
                                           atom.occupancy,
                                           atom.tempfactor,
                                           atom.element.symbol,
                                           atom.charge,
                                           ))
        self.f.write('ENDMDL\n')
        
    def write(self, trajectory):
        if isinstance(trajectory, np.ndarray):
            trajectory = mdio.base.Trajectory(trajectory)
        for i in range(len(trajectory)):
            self.write_frame(trajectory.frame(i))

    def close(self):
        self.f.close()
        
    def __enter__(self):
        return self
    
    def __exit__(self, *args):
        self.close()

def pdb_open(filename, mode='r', top=None, selection=None, use_variants=False):
    """
    Open a PDB file.
    """
    if not mode in ["r", "w"]:
        raise ValueError('Error: mode must be "r" or "w".')
    if mode == 'r':
        return PDBFileReader(filename, top=None, selection=selection)
    else:
        return PDBFileWriter(filename, top=top, use_variants=use_variants)



import numpy as np
import warnings
from  mdio import rmsd_utils
from os.path import splitext, isfile
#from mendeleev import element as ele
#from scipy.spatial.distance import cdist
from mdio.noscipy import cdist
import mdio.dcdio 
import mdio.ncio 
import mdio.xtcio 
import mdio.pdbio 
import mdio.groio
import mdio.rstio
import mdio.mdcrdio
import mdio.mdioio
import mdio.mdtio
import mdio.openmmio
from mdio.utilities import la2v, v2la, parse_selection, Imager, selection_to_indices
from mdio.data import protein_rnames, nucleic_rnames

class Frame(object):
    """
    A frame of trajectory data.
    
    """
    def __init__(self, xyz, topology=None, box=None, time=0.0, precision=1000, timestep = 1.0, units='nanometers'):
        if units == 'nanometers':
            self.units = 1.0
        elif units == 'angstroms':
            self.units = 0.1

        xyz = np.array(xyz, dtype=np.float32) * self.units
        self.topology = topology
        if len(xyz.shape) != 2:
            raise TypeError('Error - crds must be a [N,3] array.')
        if xyz.shape[1] != 3:
            raise TypeError('Error - crds must be a [N,3] array.')
        self.n_atoms = xyz.shape[0]
        self.xyz = xyz
        if box is not None:
            box = np.array(box, dtype=np.float32) * self.units
            if box.min() == 0.0 and box.max() == 0.0:
                box = None
            elif len(box.shape) == 1 and len(box) == 6:
                tbox = np.zeros((3,3), dtype=np.float32)
                tbox[0, 0] = box[0]
                tbox[1, 0] = box[1]
                tbox[1, 1] = box[2]
                tbox[2, 0] = box[3]
                tbox[2, 1] = box[4]
                tbox[2, 2] = box[5]
                box = tbox
            elif box.shape != (3,3):
                raise ValueError('Error - unrecognised box data {}'.format(box))
        self.unitcell_vectors = box
        if self.unitcell_vectors is not None:
            self.unitcell_lengths, self.unitcell_angles = v2la(self.unitcell_vectors)
        else:
            self.unitcell_lengths = None
            self.unitcell_angles = None
        self.time = float(time)
        self.precision = int(precision)
        self.timestep = timestep

    def __str__(self):
        if self.unitcell_vectors is not None:
            return "mdio.Frame with {} atoms and box info.".format(self.n_atoms)
        else:
            return "mdio.Frame with {} atoms.".format(self.n_atoms)

    def select(self, selection):
        """
        Create a new Frame from selected atoms.
        """
        atom_indices = selection_to_indices(selection, self.topology)
        if self.topology is not None:
            top = self.topology.subset(atom_indices)
        else:
            top = None
        return Frame(self.xyz[atom_indices], top, self.unitcell_vectors, self.time)

    def rmsd_from(self, frame, weights=None):
        """
        The RMSD of this Frame from a reference Frame.
        """
        if isinstance(frame, Trajectory):
            frame = frame[0]
        elif not isinstance(frame, Frame):
            raise TypeError('Error - argument must be a Frame or Trajectory')

        if frame.xyz.shape != self.xyz.shape:
            raise ValueError("Error - reference structure has {} atoms but frame has {} atoms.".format(frame.xyz.shape[0], self.xyz.shape[0]))
        return rmsd_utils.kabsch_rmsd(self.xyz, frame.xyz, weights)

    def fitted_to(self, frame, weights=None):
        """
        Returns a copy of the frame least-squares fitted to the reference.
        """
        if isinstance(frame, Trajectory):
            frame = frame[0]
        elif not isinstance(frame, Frame):
            raise TypeError('Error - argument must be a Frame or Trajectory')

        if frame.xyz.shape != self.xyz.shape:
            raise ValueError("Error - reference structure has {} atoms but frame has {} atoms.".format(frame.xyz.shape[0], self.xyz.shape[0]))

        xyz = rmsd_utils.kabsch_fit(self.xyz, frame.xyz, weights)
        return Frame(xyz, self.topology, None, self.time)

    def packed_around(self, selection=None):
        """
        Pack the coordinates in a frame into the periodic box.
        """
        if self.unitcell_vectors is None:
            print('DEBUG: No PBC info.')
            return self
        im = Imager(self.unitcell_vectors)
        if selection is None:
            atom_indices = range(self.n_atoms)
        else:
            atom_indices = selection_to_indices(selection, self.topology)

        if self.topology is not None:
            atomlists = [c.atom_indices for c in self.topology.chains]
        else:
            atomlists = [range(self.n_atoms)]
        f = self.make_whole(atomlists)
        focus = f.xyz[atom_indices].mean(axis=0)
        for atomlist in atomlists:
            cog = (f.xyz[atomlist].max(axis=0) + f.xyz[atomlist].min(axis=0)) / 2
            dx = cog - focus
            shift = dx - im.image(dx)
            f.xyz[atomlist] -= shift

        return f

    def make_whole(self, atomlists):
        """
        Image atoms to make molecules whole.
        """
        if self.unitcell_vectors is None:
            return self

        im = Imager(self.unitcell_vectors)
        xyz = self.xyz.copy()
        for atomlist in atomlists:
            for i in range(1, len(atomlist)):
                dx = xyz[atomlist[i]] - xyz[atomlist[i-1]]
                dx = im.image(dx)
                xyz[atomlist[i]] = dx + xyz[atomlist[i-1]]

        return Frame(xyz, self.topology, self.unitcell_vectors, self.time)

    def save(self, filename):
        ext = splitext(filename)[1]
        if ext in [".nc", ".ncdf"]:
            opener = mdio.ncio.nc_open
        elif ext in [".ncrst"]:
            opener = mdio.ncio.ncrst_open
        elif ext in [".dcd"]:
            opener = mdio.dcdio.dcd_open
        elif ext in [".xtc"]:
            opener = mdio.xtcio.xtc_open
        elif ext in [".pdb"]:
            opener = mdio.pdbio.pdb_open
        elif ext in [".gro"]:
            opener = mdio.groio.gro_open
        elif ext in [".rst", ".rst7"]:
            opener = mdio.rstio.rst_open
        elif ext in [".mdcrd"]:
            opener = mdio.mdcrdio.mdcrd_open
        else:
            raise TypeError('Error - unrecognised file extension ({})'.format(ext))
        with opener(filename, "w") as f:
            f.write_frame(self)


# Trajectory-level objects for mdio

class Trajectory(object):
    """
    A series of Frames.
    """
    def __init__(self, data, top=None):
        if isinstance(data, str):
            frames = mdio.load(data).frames()

        elif isinstance(data, Frame):
            frames = [data]
        elif isinstance(data, list):
            if not isinstance(data[0], Frame):
                raise TypeError('Error - argument must be a frame or list of frames.')
            frames = data
        elif isinstance(data, np.ndarray):
            s = data.shape
            if len(s) < 2 or len(s) > 3:
                raise TypeError('Error - argument must be a [natoms, 3] or [nframes, natoms, 3] array')
            if s[-1] != 3:
                raise TypeError('Error - argument must be a [natoms, 3] or [nframes, natoms, 3] array')
            if len(s) == 2:
                data = [data]
            frames = []
            for i, x in enumerate(data):
                frames.append(Frame(x, time=i)) 
        else:
            raise TypeError('Error - unsupported data type ({}) for initialisation.'.format(type(data)))

        self.xyz = np.array([frames[0].xyz])
        self.unitcell_vectors = np.array([frames[0].unitcell_vectors])
        self.time = np.array([frames[0].time])
        if len(frames) > 1:
            self.append(frames[1:])
        if top is not None:
            if isinstance(top, Topology):
                self.topology = top
            else:
                self.topology = Topology(top)
        else:
            if frames[0].topology is not None:
                self.topology = frames[0].topology
            else:
                self.topology = None
        self.top = self.topology
        self.comm = None

    def __str__(self):
        if self.unitcell_vectors[0] is None:
            return "mdio.Trajectory with {} frames, and {} atoms.".format(len(self.xyz), self.n_atoms)
        else:
            return "mdio.Trajectory with {} frames, {} atoms and box info.".format(len(self.xyz), self.n_atoms)

    def __len__(self):
        """
        Length of the trajectory.
        """
        return len(self.xyz)

    def __getitem__(self, key):
        """
        Returns a sub-Trajectory.
        """
        xyz = self.xyz[key]
        unitcell_vectors = self.unitcell_vectors[key]
        time = self.time[key]
        if isinstance(key, int):
            return Trajectory(Frame(xyz, self.topology, unitcell_vectors, time), top=self.topology)
        else:
            return Trajectory([Frame(xyz[i], self.topology, unitcell_vectors[i], time[i]) for i in range(len(xyz))], top=self.topology)

    def __add__(self, other):
        if not isinstance(other, Trajectory):
            raise NotImplementedError('Error - can only add a Trajectory to a Trajectory.')
        result = Trajectory(self.frames())
        result.append(other.frames())
        return result

    def __iadd__(self, other):
        if not isinstance(other, Trajectory):
            raise NotImplementedError('Error - can only add a Trajectory to a Trajectory.')
        self.append(other.frames())
        return self

    @property
    def n_frames(self):
        return self.xyz.shape[0]

    @property
    def n_atoms(self):
        return self.xyz.shape[1]

    def append(self, data):
        """
        Append exra data to a Trajectory

        Data may be a single Frame or a list of Frames.
.
        """
        if isinstance(data, Frame):
            frames = [data]
        elif isinstance(data, list):
            if all([isinstance(d, Frame) for d in data]):
                frames = data
            else:
                raise TypeError('Error - argument must be a frame or list of frames')
        else:
            raise TypeError('Error - argument must be a frame or list of frames.')
        xyz = []
        unitcell_vectors = []
        time = []
        for frame in frames:
            if frame.xyz.shape != self.xyz[0].shape:
                raise ValueError('Error - all frames must contain the same number of atoms.')
            if (frame.unitcell_vectors is None and self.unitcell_vectors[0] is not None) or (frame.unitcell_vectors is not None and self.unitcell_vectors[0] is None):
                raise ValueError('Error - mixing frames with and without box info.')
            xyz.append(frame.xyz)
            unitcell_vectors.append(frame.unitcell_vectors)
            time.append(frame.time)

        self.xyz = np.vstack((self.xyz, xyz))
        self.time = np.concatenate((self.time, time))
        if self.unitcell_vectors[0] is None:
            self.unitcell_vectors = np.concatenate((self.unitcell_vectors, unitcell_vectors))
        else:
            self.unitcell_vectors = np.vstack((self.unitcell_vectors, unitcell_vectors))

    def frame(self, index):
        """
        Returns one frame from the trajectory.
        """
        return Frame(self.xyz[index], self.topology, self.unitcell_vectors[index], self.time[index])
 
    def frames(self):
        """
        Returns the trajectory as a list of frames.
        """
        return [self.frame(i) for i in range(len(self))]

    def select(self, selection):
        """
        Create a new Trajectory from selected atoms.
        """
        atom_indices = selection_to_indices(selection, self.topology)
        frames = []
        for i in range(len(self.xyz)):
            frames.append(Frame(self.xyz[i][atom_indices], None, self.unitcell_vectors[i], self.time[i]))
        if self.topology is not None:
            top = self.topology.subset(atom_indices)
        else:
            top = None
        return Trajectory(frames, top=top)

    def rmsd_from(self, frame, weights=None):
        """
        The RMSD of each Frame from a reference Frame.
        """
        if isinstance(frame, Trajectory):
            frame = frame.frame(0)
        elif not isinstance(frame, Frame):
            raise TypeError('Error - argument must be a Frame or Trajectory')

        if frame.xyz.shape != self.xyz[0].shape:
            raise ValueError("Error - reference structure has {} atoms but trajectory has {} atoms.".format(frame.xyz.shape[0], self.xyz.shape[0]))
        return [rmsd_utils.kabsch_rmsd(xyz, frame.xyz, weights) for xyz in self.xyz]

    def fitted_to(self, frame, weights=None):
        """
        Returns a copy of the trajectory least-squares fitted to the reference.
        """
        if isinstance(frame, Trajectory):
            frame = frame.frame(0)
        elif not isinstance(frame, Frame):
            raise TypeError('Error - argument must be a Frame or Trajectory')

        if frame.xyz.shape != self.xyz[0].shape:
            raise ValueError("Error - reference structure has {} atoms but trajectory has {} atoms.".format(frame.xyz.shape[0], self.xyz.shape[0]))

        xyz = [rmsd_utils.kabsch_fit(x, frame.xyz, weights) for x in self.xyz]
        frames = []
        for i in range(len(xyz)):
            frames.append(Frame(xyz[i], None, None, self.time[i]))
        return Trajectory(frames, top=self.topology)

    def superpose(self, reference, atom_indices=None):
        """
        In-place least-squares fit of trajectory to reference coordinates.
        """
        if atom_indices is not None:
            weights = np.zeros(self.n_atoms)
            weights[atom_indices] = 1.0
        else:
            weights = None
        new = self.fitted_to(reference, weights=weights)
        self.xyz = new.xyz
        return self

    def packed_around(self, selection=None):
        """
        Pack the coordinates in a trajectory into the periodic box.

        The centre of geometry of the selected atoms is the focus
        for the imaging; if none are given the box centre is used.
t
        """
        frames = []
        for frame in self.frames():
            frames.append(frame.packed_around(selection))
        return Trajectory(frames, top=self.topology)

    def make_whole(self):
        """
        Returns a copy of the trajectory corrected for PBC artifacts.
        """
        if self.unitcell_vectors is None:
            return self
        frames = []
        if self.topology is None:
            atomlists = [range(self.n_atoms)]
        else:
            atomlists = [c.atom_indices for c in self.topology.chains]
        for frame in self.frames():
            frames.append(frame.make_whole(atomlists))
        return Trajectory(frames, top=self.topology)

    def mean_structure(self):
        """
        Returns a Frame with the mean structure.
        """
        xyz = self.xyz.mean(axis=0)
        time = np.array(self.time).mean()
        if self.unitcell_vectors[0] is not None:
            unitcell_vectors = self.unitcell_vectors.mean(axis=0)
        else:
            unitcell_vectors = None
       
        return Frame(xyz, self.topology, unitcell_vectors, time)


    def save(self, filename, **kwargs):
        """
        Save the trajectory to a file.

        The format is deduced from the filename extension.

        """
        ext = splitext(filename)[1]
        if ext in [".nc", ".ncdf"]:
            opener = mdio.ncio.nc_open
        elif ext in [".ncrst"]:
            opener = mdio.ncio.ncrst_open
        elif ext in [".dcd"]:
            opener = mdio.dcdio.dcd_open
        elif ext in [".xtc"]:
            opener = mdio.xtcio.xtc_open
        elif ext in [".pdb"]:
            opener = mdio.pdbio.pdb_open
        elif ext in [".rst", "rst7"]:
            opener = mdio.rstio.rst_open
        elif ext in [".mdcrd"]:
            opener = mdio.mdcrdio.mdcrd_open
        elif ext in [".gro"]:
            opener = mdio.groio.gro_open
        else:
            raise TypeError('Error - unrecognised file extension ({})'.format(ext))
        with opener(filename, "w", top=self.topology, **kwargs) as f:
            f.write(self)

    def save_pdb(self, filename, use_variants=False):
        """
        Included explicitly for nglviewer compatibility.
        """
        opener = mdio.pdbio.pdb_open
        with opener(filename, "w", top=self.topology, use_variants=use_variants) as f:
            f.write(self)


class Topology(object):

    """
    A basic PDB based topology object from which selections can be turned into lists of atom indices.
    
    """
    def __init__(self, source):
        if isinstance(source, mdio.base.Topology):
            self.atomlist = source.atomlist
        elif isinstance(source, str):
            ttop = mdio.load(source)
            self.atomlist = ttop.topology.atomlist
            if self.atomlist is None:
                raise IOError('Error getting topology from {}.'.format(source))
        elif isinstance(source, list) and isinstance(source[0], dict):
            self.atomlist = source
        else:
            raise ValueError('Error initialising topology with given data type {}.'.format(type(source)))
             
        self.n_atoms = len(self.atomlist)
        self.n_residues = 0
        self.n_chains = 0
        self.bonds = []
        self.residuelist = []
        self.chainlist = []
        rcname = None
        cname = None
        prot_nuc_rnames = protein_rnames + nucleic_rnames
        
        for a in self.atomlist:
            reschain = str(a['resSeq']) + '_' + a['chainID']
            chainid = a['chainID']
            if cname != chainid or (reschain != rcname and not a['resName'].strip() in prot_nuc_rnames):
                self.n_chains += 1
                cname = chainid
                c = {}
                c['chainID'] = chainid
                c['index'] = self.n_chains - 1
                c['atom_indices'] = []
                self.chainlist.append(c)

            if reschain != rcname:
                self.n_residues += 1
                rcname = reschain
                r = {}
                r['index'] = self.n_residues - 1
                r['chainIndx'] = self.n_chains - 1
                r['resSeq'] = a['resSeq']
                r['chainID'] = a['chainID']
                r['resName'] = a['resName']
                r['iCode'] = a['iCode']
                r['variant'] = None
                self.residuelist.append(r)

            a['resIndx'] = self.n_residues - 1
            a['chainIndx'] = self.n_chains - 1

        for i, a in enumerate(self.atomlist):
            a['mass'] = self.atom(i).element.mass
            a['element'] = self.atom(i).element.symbol
            self.chainlist[a['chainIndx']]['atom_indices'].append(i)
        

    def __str__(self):
        return 'mdio.Topology with {} atoms, {} residues, {} chains and {} bonds.'.format(self.n_atoms, self.n_residues, self.n_chains, len(self.bonds))


    def get_bonds(self, cdist, xyz, rscale=1.0, selection1=None, selection2=None):
        """
        Find the bonds between the selected sets of atoms.

        The argument cdist is the function used to calculate inter-atom
        distances. This might be scipy.spatial.distance.cdist, but might
        be mdio.utilities.Imager.cdist which handles PBC.

        """
        if selection1 is None:
            atom_indices1 = range(self.n_atoms)
        else:
            if isinstance(selection1, str):
                atom_indices1 = self.select(selection1)
            else:
                try:
                    atom_indices1 = list(selection1)
                except:
                    raise TypeError('Error - selection must be a string or list.')
        if selection2 is None:
            atom_indices2 = range(self.n_atoms)
        else:
            if isinstance(selection2, str):
                atom_indices2 = self.select(selection2)
            else:
                try:
                    atom_indices2 = list(selection2)
                except:
                    raise TypeError('Error - selection must be a string or list.')
        d = cdist(xyz[atom_indices1], xyz[atom_indices2])
        bonds = []
        for i, iindx in enumerate(atom_indices1):
            ri = self.atom(iindx).element.radius
            for j, jindx in enumerate(atom_indices2):
               rj = self.atom(jindx).element.radius
               dmax = (ri + rj) * rscale
               if d[i, j] < dmax:
                   bonds.append([iindx, jindx])

        return bonds

    def set_bonds(self, xyz, unitcell_vectors=None, rscale=1.1):
        """
        Set the bonds attribute of the topology.
 
        For speed, the method assumes that bonds within protein 
        (or nucleic) regions may be inter-residue, but bonds within 
        hetero components must be intra-residue.

        The parameter rscale scales the covalent radii values used
        to define threshold distances for bonds. The values from the
        mendeleev package are tight, scaling by about 1.1 usually finds
        all the true bonds in a decent structure, without artifacts.

        """
        if unitcell_vectors is None:
            distfunc = cdist
        else:
            im = Imager(unitcell_vectors)
            distfunc = im.cdist

        atoms = self.select('protein or nucleic')
        bonds = []
        for residue in self.residues:
            indices = [atom.index for atom in residue.atoms]
            new_bonds = self.get_bonds(distfunc, xyz, rscale=rscale, selection1=indices, selection2=indices)
            for bond in new_bonds:
                if bond[1] > bond[0]:
                    bonds.append(bond) 

        for chain in self.chains:
            rlist = list(chain.residues)
            for ir in range(len(rlist) - 1):
                residue_i = rlist[ir]
                if list(residue_i.atoms)[0].index in atoms:
                    residue_j = rlist[ir+1]
                    indices_i = [atom.index for atom in residue_i.atoms]
                    indices_j = [atom.index for atom in residue_j.atoms]
                    new_bonds = self.get_bonds(distfunc, xyz, rscale=rscale, selection1=indices_i, selection2=indices_j)
                    bonds += new_bonds

        self.bonds = bonds

    def atom(self, index):
        return Atom(index, self)

    @property
    def atoms(self):
        indx = 0
        while indx < self.n_atoms:
            yield Atom(indx, self)
            indx += 1

    def residue(self, index):
        return Residue(index, self)

    @property
    def residues(self):
        resindx = 0
        while resindx < self.n_residues:
            yield Residue(resindx, self)
            resindx += 1

    def chain(self, index):
        return Chain(index, self)

    @property
    def chains(self):
        chainindx = 0
        while chainindx < self.n_chains:
            yield Chain(chainindx, self)
            chainindx += 1

    def select(self, expression):
        atom_indices = []
        try:
            selection_expression = parse_selection(expression)
        except:
            raise ValueError('Error: cannot parse selection expression "{}".'.format(expression))
        for atom in self.atoms:
            if eval(selection_expression):
                atom_indices.append(atom.index)
        return atom_indices

    def subset(self, selection):
        atom_indices = selection_to_indices(selection, self)
        alist = [a.copy() for a in self.atomlist if a['index'] in atom_indices]
        return Topology(alist)

    def set_residue_variants(self, variants_list):
        if len(variants_list) != self.n_residues:
            raise ValueError('Error - variants list must have {} elements.'.format(self.n_residues))
        for i, v in enumerate(variants_list):
            self.residue(i).variant = v

    @classmethod
    def from_dataframe(cls, atoms, bonds=None):
        if len(atoms) == 0:
            raise ValueError('Error - no atoms.')
        atomlist = atoms.to_dict(orient='records')
        return mdio.base.Topology(atomlist)
    
class Element(object):
    element_masses = {}
    element_radii = {}

    def __init__(self, symbol='X', mass=1.0, radius=1.0):
        self.symbol = symbol
        self.mass = mass
        self.radius = radius

    @classmethod
    def from_symbol(self, symbol):
        # covalent radii in picometers
        eledict = {
            'H': [1.007, 37],
            'C': [12.01, 77],
            'N': [14.01, 75],
            'O': [16.00, 73],
            'F': [19.00, 71],
            'P': [30.97, 106],
            'S': [32.07, 102],
            'Mg': [24.30, 141],
            'Ca': [40.08, 176],
            'Cl': [35.45, 99],
            'Br': [79.90, 114],
            'I': [125.91, 133]}

        if not symbol in self.element_masses:
            try:
                e = eledict[symbol.strip()]
            except KeyError:
                warnings.warn('Warning: unrecognised element symbol {}'.format(symbol))
                e = eledict['C']
            self.element_masses[symbol] = e[0]
            self.element_radii[symbol] = e[1] * 0.001 # to nm
        return Element(symbol, self.element_masses[symbol], self.element_radii[symbol])
            
class Atom(object):
    def __init__(self, index, topology):
        self.index = index
        self.topology = topology
        resindx = self.topology.atomlist[index]['resIndx']
        self.residue = self.topology.residue(resindx)
        chainindx = self.topology.atomlist[index]['chainIndx']
        self.chain = self.topology.chain(chainindx)
        self.altloc = self.topology.atomlist[index]['altLoc']
        
    @property
    def serial(self):
        return self.topology.atomlist[self.index]['serial']

    @serial.setter
    def serial(self, value):
        self.topology.atomlist[self.index]['serial'] = value

    @property
    def name(self):
        return self.topology.atomlist[self.index]['name']

    @name.setter
    def name(self, value):
        self.topology.atomlist[self.index]['name'] = value

    @property
    def occupancy(self):
        return self.topology.atomlist[self.index]['occupancy']

    @occupancy.setter
    def occupancy(self, value):
        self.topology.atomlist[self.index]['occupancy'] = value

    @property
    def tempfactor(self):
        return self.topology.atomlist[self.index]['tempFactor']

    @tempfactor.setter
    def tempfactor(self, value):
        self.topology.atomlist[self.index]['tempFactor'] = value

    @property
    def element(self):
        return Element.from_symbol(self.topology.atomlist[self.index]['element'])

    @element.setter
    def element(self, value):
        self.topology.atomlist[self.index]['element'] = value.rjust(2)
    
    @property
    def charge(self):
        return self.topology.atomlist[self.index]['charge']

    @charge.setter
    def charge(self, value):
        self.topology.atomlist[self.index]['charge'] = value


            
class Residue(object):
    def __init__(self, resindx, topology):
        self.resindx = resindx
        self.topology = topology
        self.name = self.topology.residuelist[resindx]['resName']
        chainindx = self.topology.residuelist[resindx]['chainIndx']
        self.chain = self.topology.chain(chainindx)
        self.icode = self.topology.residuelist[resindx]['iCode']

    @property
    def variant(self):
        return self.topology.residuelist[self.resindx]['variant']

    @variant.setter
    def variant(self, value):
        self.topology.residuelist[self.resindx]['variant'] = value

    @property
    def resSeq(self):
        return self.topology.residuelist[self.resindx]['resSeq']

    @resSeq.setter
    def resSeq(self, value):
        self.topology.residuelist[self.resindx]['resSeq'] = value
    
    @property
    def name(self):
        return self.topology.residuelist[self.resindx]['resName']

    @name.setter
    def name(self, value):
        self.topology.residuelist[self.resindx]['resName'] = value
    
    @property
    def atoms(self):
        indx = 0
        while indx < self.topology.n_atoms:
            if self.topology.atomlist[indx]['resIndx'] == self.resindx:
                yield Atom(indx, self.topology)
                indx += 1
            else:
                indx += 1

class Chain(object):
    def __init__(self, chainindx, topology):
        self.index = chainindx
        self.topology = topology

    @property
    def chainid(self):
        return self.topology.chainlist[self.index]['chainID']

    @chainid.setter
    def chainid(self, value):
        self.topology.chainlist[self.index]['chainID'] = value

    @property
    def residues(self):
        resindx = 0
        while resindx < self.topology.n_residues:
            if self.topology.residuelist[resindx]['chainIndx'] == self.index:
                yield Residue(resindx, self.topology)
                resindx += 1
            else:
                resindx += 1

    @property
    def atom_indices(self):
        return self.topology.chainlist[self.index]['atom_indices']


def mpi_load(filenames, top=None, selection=None, comm=None):
    """
    MPI-powered file loader
    """
    if comm is None or isinstance(filenames, str):
        traj = load(filenames, top=top, selection=selection)
        traj.comm = comm
        return traj
    else:
        size = comm.Get_size()
        rank = comm.Get_rank()
    if size == 1:
        return load(filenames, top=top, selection=selection)
    nfiles = len(filenames)
    offsets = np.rint(np.linspace(0, nfiles, size, endpoint=False)).astype(np.int32)
    offsets = np.append(offsets, [nfiles])

    i = offsets[rank]
    j = offsets[rank + 1]
    if j > i:
        t = mdio.load(filenames[i:j], top, selection=selection)
    else:
        t = mdio.load_frame(filenames[0], 0, top=top, selection=selection)
        t.xyz = t.xyz[0:0]
        t.time = t.time[0:0]
        if t.unitcell_vectors is not None:
            t.unitcell_vectors = t.unitcell_vectors[0:0]

    sendcounts = np.array(comm.gather(t.n_frames, root=0))
    sendcounts = comm.bcast(sendcounts, root=0)

    n_atoms = t.n_atoms
    tcounts = sendcounts
    xcounts = sendcounts * n_atoms * 3
    bcounts = sendcounts * 9
    tot_frames = sendcounts.sum()
    has_box = t.unitcell_vectors is not None

    xyz = np.empty((tot_frames, n_atoms, 3), dtype=t.xyz.dtype)
    time = np.empty(tot_frames, dtype=t.time.dtype)
    if has_box:
        unitcell_vectors = np.empty((tot_frames, 3, 3), dtype=t.unitcell_vectors.dtype)
    else:
        unitcell_vectors = None

    comm.Gatherv(t.xyz, (xyz, xcounts), root=0)
    comm.Gatherv(t.time, (time, tcounts), root=0)
    comm.Bcast(xyz, root=0)
    comm.Bcast(time, root=0)
    
    if has_box:
        comm.Gatherv(t.unitcell_vectors, (unitcell_vectors, bcounts),  root=0)
        comm.Bcast(unitcell_vectors, root=0)

    t_all = mdio.Trajectory(xyz, t.topology) 
    t_all.time = time
    t_all.unitcell_vectors = unitcell_vectors
    t_all.comm = comm
    return t_all

def slice_parse(filename):
    """
    Split a filename into real name and possible [start:stop:step] components
    """
    def int_or_none(c):
        try:
            result = int(c)
        except:
            result = None
        return result

    if not isinstance(filename, str):
        return filename, None

    if '[' in filename:
        if not ']' in filename:
            raise ValueError('Error in filename {}.'.format(filename))
        i = filename.find('[')
        j = filename.find(']')
        indices = filename[i+1:j]
        filename = filename[:i]
        key = slice(*[int_or_none(c) for c in indices.split(':')])
    else:
        key = None
    return filename, key

def get_opener(filename, top=None):
    """
    Returns an opener for filename.
    """
    try:
        ok = isfile(filename)
    except:
        ok = True
    if not ok:
        raise IOError('Error: file "{}" does not exist.'.format(filename))
    openers = [mdio.ncio.nc_open, 
               mdio.ncio.ncrst_open, 
               mdio.dcdio.dcd_open, 
               mdio.xtcio.xtc_open, 
               mdio.groio.gro_open, 
               mdio.rstio.rst_open, 
               mdio.mdtio.mdt_open, 
               mdio.mdioio.mdio_open, 
               mdio.openmmio.openmm_open, 
               mdio.pdbio.pdb_open, 
               mdio.mdcrdio.mdcrd_open]
    for opener in openers:
        if opener == mdio.mdcrdio.mdcrd_open:
            tmptop = top
        else:
            tmptop = None
        try:
            with opener(filename, top=tmptop) as f:
                frame = f.read_frame()
            good_opener = opener
            success = True
        except:
            success = False
        if success:
            break
    if not success:
        return None
    else:
        return good_opener

def load(filenames, top=None, selection=None):
    """
    Format-detecting file loader
    """

    tlist = []
    if not isinstance(filenames, list):
        filenames = [filenames,]
    for i, filename in enumerate(filenames):
        filename, key = slice_parse(filename)
        opener = get_opener(filename, top=top)
        if opener is None:
            raise TypeError('Error - {} does not have a recognised file format'.format(filename))
        with opener(filename, top=top, selection=selection) as f:
            if key is None:
                tlist.append(f.read())
            else:
                tlist.append(f.read()[key])
    result = tlist[0]
    for t in tlist[1:]:
        result += t
    return result

def load_frame(filename, index, top=None, selection=None):
    """
    Format-detecting file loader for a single frame.
    """
    opener = get_opener(filename, top=top)
    if opener is None:
        raise TypeError('Error - {} does not have a recognised file format'.format(filename))
    i = 0
    with opener(filename, top=top, selection=selection) as f:
        while i <= index:
            frame = f.read_frame()
            i += 1
    return Trajectory(frame, top=top)

def mdopen(filename, mode="r", top=None, selection=None):
    """
    Format-agnostic open routines.
    """
    if not mode in ["r", "w"]:
        raise ValueError('Error - mode must be "r" or "w"')

    if mode == "r":
        f = open(filename, 'rb')
        f.close()
        if top is not None:
            f = open(top, 'rb')
            f.close()
        opener = get_opener(filename, top=top)
        if opener is None:
            raise TypeError('Error - {} does not have a recognised file format'.format(filename))
        return opener(filename, top=top, selection=selection)

    else:
        ext = splitext(filename)[1]
        if ext in [".nc", ".ncdf"]:
            opener = mdio.ncio.nc_open
        elif ext in [".ncrst"]:
            opener = mdio.ncio.ncrst_open
        elif ext in [".dcd"]:
            opener = mdio.dcdio.dcd_open
        elif ext in [".xtc"]:
            opener = mdio.xtcio.xtc_open
        elif ext in [".pdb"]:
            opener = mdio.pdbio.pdb_open
        elif ext in [".gro"]:
            opener = mdio.groio.gro_open
        elif ext in [".rst", ".rst7"]:
            opener = mdio.rstio.rst_open
        elif ext in [".mdcrd"]:
            opener = mdio.mdcrdio.mdcrd_open
        else:
            raise TypeError('Error - unrecognised file extension ({})'.format(ext))
        return opener(filename, "w", top=top)

def get_indices(pdbfile, selection):
    top = Topology(pdbfile)
    return top.select(selection)

def rmsd(traj1, traj2):
    return traj1.rmsd_from(traj2)




import mdio.base
from tempfile import NamedTemporaryFile

class OpenMMReader(object):
    def __init__(self, trajectory, top=None, selection=None):

        n_frames = trajectory.getNumFrames()
        topology = trajectory.getTopology()
        pdbfile = NamedTemporaryFile(mode='w', suffix='.pdb')
        trajectory.writeHeader(topology, pdbfile)
        for frame in range(n_frames):
            positions = trajectory.getPositions(frame=frame)
            pdbfile.write('MODEL  {:6d}\n'.format(frame))
            trajectory.writeModel(topology, positions, pdbfile)
            pdbfile.write('ENDMDL\n')
        trajectory.writeFooter(topology, pdbfile)
        pdbfile.flush()
        self.trajectory = mdio.base.load(pdbfile.name, 
                                         selection=selection)
        pdbfile.close()
        self.index = -1
        self.n_frames = int(self.trajectory.n_frames)
        assert self.n_frames == n_frames

    def read_frame(self):
        self.index += 1
        if self.index >= self.n_frames:
            return None

        return self.trajectory.frame(self.index)

    def read(self):
        self.index += 1
        return self.trajectory[self.index:]

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


def openmm_open(trajectory, mode='r', top=None, selection=None):
    """
    Open an OpenMM "trajectory"
    """
    if mode != 'r':
        raise IOError('Error: OpenMM trajectories can only be read')
    return OpenMMReader(trajectory, top=top,
                        selection=selection)




import mdio
import numpy as np
import pytest
try:
    import mdtraj as mdt
    has_mdt = True
except:
    has_mdt = False

try:
    from simtk.openmm.app import PDBFile
    has_openmm = True
except:
    has_openmm = False

ncfile = 'examples/test.nc'
dcdfile = 'examples/test.dcd'
xtcfile = 'examples/test.xtc'
pdbfile = 'examples/test.pdb'
pdb10file = 'examples/test10.pdb'
grofile = 'examples/test.gro'
testcrds = np.array([2.837, 4.348, 2.527])
testbox = np.array([[5.9414,  0.,     0.   ],
     [ 0.,    5.9414,  0.,   ],
     [ 0.,     0.,    5.9414]], dtype=np.float32)
testi = 10

def test_loader():
    t = mdio.load(ncfile)
    assert np.allclose(t.xyz[0,0], testcrds) is True
    assert np.allclose(t[0].unitcell_vectors, testbox) is True
    assert len(t) == testi

def test_opener():
    fin = mdio.open(ncfile)
    fout = mdio.open('tmp.dcd', 'w')
    frame = fin.read_frame()
    while frame is not None:
        fout.write_frame(frame)
        frame = fin.read_frame()
    fin.close()
    fout.close()

def test_mdt_load():
    if not has_mdt:
        pytest.skip("mdtraj not installed.")
    mdtrajectory = mdt.load(ncfile, top=pdbfile)
    t = mdio.load(mdtrajectory)
    assert t.n_frames == mdtrajectory.n_frames

def test_openmm_load():
    if not has_openmm:
        pytest.skip("openmm not installed.")
    openmm_molecule  = PDBFile(pdb10file)
    t = mdio.load(openmm_molecule)
    assert t.n_frames == openmm_molecule.getNumFrames()

def test_consistency():
    finA = mdio.open(ncfile, 'r')
    finB = mdio.open('tmp.dcd', 'r')
    frameA = finA.read_frame()
    frameB = finB.read_frame()
    while frameA is not None:
        assert np.allclose(frameA.xyz, frameB.xyz) is True
        frameA = finA.read_frame()
        frameB = finB.read_frame()
    finA.close()
    finB.close()

def test_multiopen():
    t = mdio.load([ncfile, dcdfile])
    assert len(t) == 20

def test_with_top():
    t = mdio.load(ncfile, top=pdbfile)
    assert t.topology is not None

def test_with_top_and_indices():
    t = mdio.load(ncfile, top=pdbfile, selection=range(10))
    assert t.n_atoms == 10
    sel = t.topology.select('name CA')
    assert len(sel) == 1

def test_with_selection_1():
    t = mdio.load(ncfile, top=pdbfile, selection='name CA')
    assert t.n_atoms == 58

def test_with_selection_2():
    t = mdio.load(ncfile, top=pdbfile, selection='not residue 1 to 3')
    assert t.n_atoms == 840

def test_with_selection_3():
    t = mdio.load(xtcfile, top=grofile, selection='name CA')
    assert t.n_atoms == 58

def test_with_selection_4():
    t = mdio.load(dcdfile, top=grofile, selection='not residue 1 to 3')
    assert t.n_atoms == 840

def test_with_selection_5():
    t = mdio.load(dcdfile, top=grofile, selection='element N')
    assert t.n_atoms == 84

def test_with_selection_6():
    t = mdio.load(dcdfile, top=grofile, selection='mass > 2.0')
    assert t.n_atoms == 454

def test_with_slice_2():
    t = mdio.load(dcdfile + '[3:10:3]', top=grofile, selection='mass > 2.0')
    assert t.n_frames == 3

def test_with_slice_1():
    t = mdio.load(dcdfile + '[:5]', top=grofile, selection='mass > 2.0')
    assert t.n_frames == 5

def test_make_whole():
    t = mdio.load(pdbfile)
    t2 = t.packed_around(selection='index 1')
    assert t2.rmsd_from(t)[0] < 0.01
    t3 = t2.make_whole()
    assert t3.rmsd_from(t)[0] < 0.01



import mdio
from mdio import Frame, Trajectory, Topology
from mdio.base import  get_indices
import numpy as np

#data = np.random.rand(13,56,3)
data = np.arange(13 * 56 * 3).reshape((13, 56, 3))
pdbfile = 'examples/test.pdb'

def test_frame_init():
    f = Frame(data[0])

def test_frame_select():
    atom_indices = range(5)
    f = Frame(data[0])
    fs = f.select(atom_indices)
    assert fs.xyz.shape[0] == 5

def test_frame_rmsd_from():
    f = Frame(data[0])
    f2 = Frame(data[0])
    f2.xyz[:, 0] = f.xyz[:, 1]
    f2.xyz[:, 1] = f.xyz[:, 2]
    f2.xyz[:, 2] = f.xyz[:, 0]
    assert f.rmsd_from(f2) < 0.01

def test_frame_fitted_to():
    f = Frame(data[0])
    f2 = Frame(data[0])
    f2.xyz[:, 0] = f.xyz[:, 1]
    f2.xyz[:, 1] = f.xyz[:, 2]
    f2.xyz[:, 2] = f.xyz[:, 0]
    f3 = f.fitted_to(f2)
    assert np.allclose(f3.xyz, f2.xyz, atol=1e-05)

def test_frame_packed_around():
    xyz = data[0]
    extent = xyz.max(axis=0) - xyz.min(axis=0)
    extent = extent + 1.0
    unitcell_vectors = np.array([[extent[0], 0, 0], [0, extent[1], 0], [0, 0, extent[2]]])
    xyz[0] = xyz.mean(axis=0)
    f = Frame(xyz, box=unitcell_vectors)
    f2 = f.packed_around([0])
    assert  np.allclose(f2.xyz, f.xyz)
    assert f2.rmsd_from(f) < 0.01

def test_traj_init_from_frame():
    t = Trajectory(Frame(data[0]))

def test_traj_init_from_frames():
    frames = [Frame(x) for x in data]
    t = Trajectory(frames)

def test_traj_init_from_array():
    t1 = Trajectory(data[0])
    t2 = Trajectory(data)

def test_traj_init_from_file():
    t2 = Trajectory(data)
    t3 = Trajectory(pdbfile)

def test_select():
    inds = get_indices('examples/test.pdb', 'name CA and residue 2 to 3')
    assert inds == [36, 42]

def test_bonding():
    top = Topology(pdbfile)
    t = Trajectory(pdbfile, top=top)
    
    assert len(t.topology.bonds) == 0
    t.topology.set_bonds(t.xyz[0], t.unitcell_vectors[0])
    assert len(t.topology.bonds) == 903



import numpy as np
import sys
#from scipy.spatial.distance import cdist
from mdio.noscipy import cdist
from pyparsing import Keyword, Word, Combine, Literal, Forward, Optional, nums, alphanums, ZeroOrMore

class TracebackSuppressor(object):
    """
    Suppress tracebacks when exceptions are raised.
    """

    def __init__(self, name):
        self.eh = sys.excepthook
        self.name = name

    def exception_handler(self, exception_type, exception, traceback):
        if self.name in exception_type.__name__:
            print("{} {}".format(exception_type.__name__, exception))
        else:
            self.eh(exception_type, exception, traceback)

    def on(self):
        sys.excepthook = self.exception_handler

    def off(self):
        sys.excepthook = self.eh

        
def process_name(token):
    return 'atom.name.strip()'

def process_index(token):
    return 'atom.index'

def process_mass(token):
    return 'atom.element.mass'

def process_symbol(token):
    return 'atom.element.symbol.strip()'

def process_resSeq(token):
    return 'atom.residue.resSeq'

def process_resid(token):
    return 'atom.residue.resindx'

def process_resname(token):
    return 'atom.residue.name.strip()'

def process_chainid(token):
    return 'atom.chain.chainid.strip()'

def process_all(token):
    return 'True'

def process_water(token):
    return 'atom.residue.name.strip() in ["SOL", "WAT", "HOH"]'

def process_protein(token):
    return 'atom.residue.name.strip() in ["ALA", "ASP", "ASN", "ASH", "CYS", \
                                  "GLY", "GLU", "GLN", "GLH", "HIS", \
                                  "HID", "HIE", "HIP", "LEU", "LYS", \
                                  "LYN", "ARG", "ARN", "MET", "PHE", \
                                  "PRO", "SER", "THR", "TYR", "VAL", \
                                  "ILE", "CYX", "CYH", "TRP"]'

def process_nucleic(token):
    return 'atom.residue.name.strip() in ["ADE", "CYT", "GUA", "THY", \
                                  "A", "C", "G", "T", "DA", "DC", "DG", "DT", \
                                  "DA5", "DC5", "DG5", "DT5", "DA3", "DC3", \
                                  "DG3", "DT3"]'
def torange(tokens):
    return 'in range({}, {})'.format(tokens[0], int(tokens[2]) + 1)

def process_string_expression(tokens):
    return '{} == {}'.format(tokens[0], tokens[1])

def process_short_num_expression(tokens):
    return '{} == {}'.format(tokens[0], tokens[1])

def quote(token):
    return '"{}"'.format(*token)

def create_parser():
    index_key = Keyword('index')
    index_key.setParseAction(process_index)
    name_key = Keyword('name')
    name_key.setParseAction(process_name)
    atom_num_key = index_key
    atom_string_key = name_key

    element_mass_key = Keyword('mass')
    element_mass_key.setParseAction(process_mass)
    element_symbol_key = Keyword('symbol')
    element_symbol_key.setParseAction(process_symbol)
    element_element_key = Keyword('element')
    element_element_key.setParseAction(process_symbol)
    element_num_key = element_mass_key
    element_string_key = element_symbol_key ^ element_element_key

    residue_residue_key = Keyword('residue')
    residue_residue_key.setParseAction(process_resSeq)
    residue_resSeq_key = Keyword('resSeq')
    residue_resSeq_key.setParseAction(process_resSeq)
    residue_resid_key = Keyword('resid')
    residue_resid_key.setParseAction(process_resid)
    residue_resname_key = Keyword('resname')
    residue_resname_key.setParseAction(process_resname)
    residue_string_key = residue_resname_key
    residue_num_key = residue_residue_key ^ residue_resSeq_key ^ residue_resid_key

    chain_chainid_key = Keyword('chainid')
    chain_chainid_key.setParseAction(process_chainid)
    chain_string_key = chain_chainid_key

    string_key = atom_string_key ^ element_string_key ^ residue_string_key ^ chain_string_key
    num_key = atom_num_key ^ element_num_key ^ residue_num_key

    string_value = Word(alphanums)
    string_value.setParseAction(quote)

    integer_value = Word(nums)
    real_value = Combine(integer_value*(None, None) + '.' + integer_value*(None, None))
    num_value = integer_value ^ real_value
    range_value = integer_value + 'to' + integer_value
    range_value.setParseAction(torange)

    string_expression = string_key + string_value
    string_expression.setParseAction(process_string_expression)

    relOp = Literal('==') ^ '>' ^ '>=' ^ '<' ^ '<=' ^ '!='
    full_num_expression = num_key + relOp + num_value
    short_num_expression = num_key + num_value
    short_num_expression.setParseAction(process_short_num_expression)
    num_expression = full_num_expression ^ short_num_expression

    range_expression = num_key + range_value

    all_expression = Keyword('all')
    all_expression.setParseAction(process_all)
    water_expression = Keyword('water')
    water_expression.setParseAction(process_water)
    nucleic_expression = Keyword('nucleic')
    nucleic_expression.setParseAction(process_nucleic)
    protein_expression = Keyword('protein')
    protein_expression.setParseAction(process_protein)

    basic_expression = string_expression ^ num_expression ^ range_expression ^ water_expression ^ protein_expression ^ all_expression ^ nucleic_expression
    expression = Forward()
    bracketed_expression = '(' + expression + ')'
    atom = basic_expression | bracketed_expression | expression
    andor = Keyword('and') ^ Keyword('or')
    expression << Optional('not') + atom + ZeroOrMore(andor + atom)
    return expression.parseString

def parse_selection(selection):
    parser = create_parser()
    ts = TracebackSuppressor('Parse')
    ts.on()
    result = ' '.join(parser(selection))
    ts.off()
    return result

def la2v(lengths, angles):
    alpha = angles[0] * np.pi / 180
    beta = angles[1] * np.pi / 180
    gamma = angles[2] * np.pi / 180

    a = np.array([lengths[0], 0.0, 0.0])
    b = np.array([lengths[1]*np.cos(gamma), lengths[1]*np.sin(gamma), 0.0])
    cx = lengths[2]*np.cos(beta)
    cy = lengths[2]*(np.cos(alpha) - np.cos(beta)*np.cos(gamma)) / np.sin(gamma)
    cz = np.sqrt(lengths[2]*lengths[2] - cx*cx - cy*cy)
    c = np.array([cx,cy,cz])

    v = np.array((a,b,c))
    # Make sure that all vector components that are _almost_ 0 are set exactly
    # to 0
    tol = 1e-6
    v[np.logical_and(v>-tol, v<tol)] = 0.0

    return v

def v2la(vectors):
    a = vectors[0]
    b = vectors[1]
    c = vectors[2]

    a_length = np.sqrt(np.sum(a*a))
    b_length = np.sqrt(np.sum(b*b))
    c_length = np.sqrt(np.sum(c*c))

    if min(a_length, b_length, c_length) == 0:
        a_length = 0.0
        b_length = 0.0
        c_length = 0.0
        alpha = 90.0
        beta = 90.0
        gamma = 90.0
    else:
        alpha = np.arccos(np.dot(b, c) / (b_length * c_length))
        beta = np.arccos(np.dot(c, a) / (c_length * a_length))
        gamma = np.arccos(np.dot(a, b) / (a_length * b_length))

        alpha = alpha * 180.0 / np.pi
        beta = beta * 180.0 / np.pi
        gamma = gamma * 180.0 / np.pi

    return [a_length, b_length, c_length], [alpha, beta, gamma]

class Imager(object):
    def __init__(self, unitcell_vectors):
        self.unitcell_vectors = unitcell_vectors
        if unitcell_vectors is not None:
            self.A = self.unitcell_vectors.T
            self.B = np.linalg.inv(self.A)

    def pack(self, xyz, centre_atom_indices=None):
        """
        Pack a set of coordinates into the periodic cell
        """
        if self.unitcell_vectors is None:
            return self
        if centre_atom_indices is not None:
            box_centre = np.matmul(self.A, [0.5, 0.5, 0.5])
            dv = box_centre - xyz[centre_atom_indices].mean(axis=0)
        else:
            dv = 0.0
        r = xyz + dv
        f = np.matmul(self.B, r.T)
        g = f - np.floor(f)
        t = np.matmul(self.A, g)
        xyz_packed = t.T - dv
        return xyz_packed

    def image(self, vector):
        """
        Return a vector imaged to be as short as possible
        """
        lmin = (vector * vector).sum()
        vopt = vector
        f = np.matmul(self.B, vector.T)
        g = f - np.floor(f)
        for a in [0, 1]:
            for b in [0, 1]:
                for c in [0, 1]:
                    gtmp = g - np.array([a, b, c])
                    t = np.matmul(self.A, gtmp)
                    l = (t * t).sum()
                    if l < lmin:
                        lmin = l
                        vopt = t
        return vopt
                
    def cdist(self, xyz1, xyz2):
        """
        cdist function with imaging.
        """
        result = cdist(xyz1, xyz2)
        ucv = self.unitcell_vectors
        for i in [-0.5, 0.5]:
            for j in [-0.5, 0.5]:
                for k in [-0.5, 0.5]:
                    shift = ucv[0] * i + ucv[1] * j + ucv[2] * k
                    xyzt1 = self.pack(xyz1 + shift)
                    xyzt2 = self.pack(xyz2 + shift)
                    result = np.stack((result, cdist(xyzt1, xyzt2))).min(axis=0)
        return result

def selection_to_indices(selection, topology):
    if selection is None:
        return None

    if isinstance(selection, str):
        if topology is not None:
            indices = topology.select(selection)
        else:
            raise TypeError('Error - selection strings require a topology.')
    else:
        try:
            indices = list(selection)
        except:
            raise TypeError('Error - selection must be a string or list.')
    return indices
